@@ -22,6 +22,7 @@
 #endif
 
 #include <Python.h>
+#include <structmember.h>
 #include <pygobject.h>
 #include <gmodule.h>
 #include <gtk/gtk.h>
@@ -42,8 +43,37 @@ static gboolean caja_python_init_python(void);
 static GArray *all_types = NULL;
 static GList *all_pyfiles = NULL;
 
+/* Caja.OperationHandle value access. */
+static PyObject *
+caja_operationhandle_get_handle(PyGBoxed *self, void *closure)
+{
+	return PyLong_FromSsize_t((Py_ssize_t) (size_t) self->boxed);
+}
 
-static inline gboolean 
+static int
+caja_operationhandle_set_handle(PyGBoxed *self, PyObject *value, void *closure)
+{
+	Py_ssize_t val = PyLong_AsSsize_t(value);
+
+	if (!PyErr_Occurred()) {
+		if (val) {
+			self->boxed = (void *) val;
+			return 0;
+		}
+		PyErr_SetString(PyExc_ValueError, "invalid operation handle value");
+	}
+	return -1;
+}
+
+static PyGetSetDef caja_operationhandle_handle = {
+	"handle",
+	(getter) caja_operationhandle_get_handle,
+	(setter) caja_operationhandle_set_handle,
+	"Operation handle value",
+	NULL
+};
+
+static inline gboolean
 np_init_pygobject(void)
 {
     PyObject *gobject = pygobject_init (PYGOBJECT_MAJOR_VERSION, PYGOBJECT_MINOR_VERSION, PYGOBJECT_MICRO_VERSION);
@@ -57,7 +87,7 @@ np_init_pygobject(void)
 }
 
 static void
-caja_python_load_file(GTypeModule *type_module, 
+caja_python_load_file(GTypeModule *type_module,
 						  const gchar *filename)
 {
 	PyObject *main_module, *main_locals, *locals, *key, *value;
@@ -75,7 +105,7 @@ caja_python_load_file(GTypeModule *type_module,
 	}
 
 	main_locals = PyModule_GetDict(main_module);
-	module = PyImport_ImportModuleEx((char *) filename, main_locals, main_locals, NULL);
+	module = PyImport_ImportModuleEx(filename, main_locals, main_locals, NULL);
 	if (!module)
 	{
 		PyErr_Print();
@@ -98,7 +128,7 @@ caja_python_load_file(GTypeModule *type_module,
 			gtype = caja_python_object_get_type(type_module, value);
 			g_array_append_val(all_types, gtype);
 
-			all_pyfiles = g_list_append(all_pyfiles, (gchar*)filename);
+			all_pyfiles = g_list_append(all_pyfiles, g_strdup(filename));
 		}
 	}
 
@@ -106,7 +136,7 @@ caja_python_load_file(GTypeModule *type_module,
 }
 
 static void
-caja_python_load_dir (GTypeModule *module, 
+caja_python_load_dir (GTypeModule *module,
 						  const char  *dirname)
 {
 	GDir *dir;
@@ -114,17 +144,17 @@ caja_python_load_dir (GTypeModule *module,
 	gboolean initialized = FALSE;
 
 	debug_enter_args("dirname=%s", dirname);
-	
+
 	dir = g_dir_open(dirname, 0, NULL);
 	if (!dir)
 		return;
-			
+
 	while ((name = g_dir_read_name(dir)))
 	{
 		if (g_str_has_suffix(name, ".py"))
 		{
 			char *modulename;
-			int len;
+			size_t len;
 
 			len = strlen(name) - 3;
 			modulename = g_new0(char, len + 1 );
@@ -133,40 +163,42 @@ caja_python_load_dir (GTypeModule *module,
 			if (!initialized)
 			{
 				PyObject *sys_path, *py_path;
-				
+
 				/* n-p python part is initialized on demand (or not
 				* at all if no extensions are found) */
 				if (!caja_python_init_python())
 				{
 					g_warning("caja_python_init_python failed");
 					g_dir_close(dir);
+					g_free(modulename);
 					break;
 				}
-				
+
 				/* sys.path.insert(0, dirname) */
 				sys_path = PySys_GetObject("path");
-				py_path = PyString_FromString(dirname);
+				py_path = PyUnicode_FromString(dirname);
 				PyList_Insert(sys_path, 0, py_path);
 				Py_DECREF(py_path);
 			}
 			caja_python_load_file(module, modulename);
+			g_free(modulename);
 		}
-	}	
+	}
 }
 
 static gboolean
 caja_python_init_python (void)
 {
-	PyObject *gi, *require_version, *args, *caja;
+	PyObject *gi, *require_version, *args, *caja, *descr;
 	GModule *libpython;
-	char *argv[] = { "caja", NULL };
+	wchar_t *argv[] = { L"caja", NULL };
 
 	if (Py_IsInitialized())
 		return TRUE;
 
 #if 0
-  	debug("g_module_open " PY_LIB_LOC "/libpython" PYTHON_VERSION "." G_MODULE_SUFFIX ".1.0");
-	libpython = g_module_open(PY_LIB_LOC "/libpython" PYTHON_VERSION "." G_MODULE_SUFFIX ".1.0", 0);
+  	debug("g_module_open " PY_LIB_LOC "/libpython" PYTHON_VERSION PYTHON_ABIFLAGS "." G_MODULE_SUFFIX ".1.0");
+	libpython = g_module_open(PY_LIB_LOC "/libpython" PYTHON_VERSION PYTHON_ABIFLAGS "." G_MODULE_SUFFIX ".1.0", 0);
 	if (!libpython)
 		g_warning("g_module_open libpython failed: %s", g_module_error());
 #endif
@@ -178,7 +210,7 @@ caja_python_init_python (void)
 		PyErr_Print();
 		return FALSE;
 	}
-	
+
 	debug("PySys_SetArgv");
 	PySys_SetArgv(1, argv);
 	if (PyErr_Occurred())
@@ -188,7 +220,7 @@ caja_python_init_python (void)
 	}
 
 	debug("Sanitize the python search path");
-	PyRun_SimpleString("import sys; sys.path = filter(None, sys.path)");
+	PyRun_SimpleString("import sys; sys.path = list(filter(None, sys.path))");
 	if (PyErr_Occurred())
 	{
 		PyErr_Print();
@@ -248,7 +280,22 @@ caja_python_init_python (void)
 	IMPORT(OperationHandle, "OperationHandle");
 
 #undef IMPORT
-	
+
+	/* Add the "handle" member to the OperationHandle type. */
+	descr = PyDescr_NewGetSet(_PyCajaOperationHandle_Type,
+							  &caja_operationhandle_handle);
+    if (!descr) {
+		PyErr_Print();
+		return FALSE;
+	}
+	if (PyDict_SetItemString(_PyCajaOperationHandle_Type->tp_dict,
+						     caja_operationhandle_handle.name, descr)) {
+		Py_DECREF(descr);
+		PyErr_Print();
+		return FALSE;
+	}
+	Py_DECREF(descr);
+
 	return TRUE;
 }
 
@@ -275,11 +322,11 @@ caja_module_initialize(GTypeModule *module)
 	caja_python_load_dir(module, DATADIR "/caja-python/extensions");
 
 	// Look in XDG_DATA_DIR, ~/.local/share/caja-python/extensions
-	user_extensions_dir = g_build_filename(g_get_user_data_dir(), 
+	user_extensions_dir = g_build_filename(g_get_user_data_dir(),
 		"caja-python", "extensions", NULL);
 	caja_python_load_dir(module, user_extensions_dir);
 }
- 
+
 void
 caja_module_shutdown(void)
 {
@@ -289,17 +336,17 @@ caja_module_shutdown(void)
 		Py_Finalize();
 
 	g_array_free(all_types, TRUE);
-	g_list_free (all_pyfiles);
+	g_list_free_full(all_pyfiles, g_free);
 }
 
 void
 caja_module_list_types(const GType **types,
-						   int          *num_types)
+                       int          *num_types)
 {
 	debug_enter();
-	
+
 	*types = (GType*)all_types->data;
-	*num_types = all_types->len;
+	*num_types = (int) all_types->len;
 }
 
 void
